from datetime import datetime
from typing import List

from cement import Controller, ex

from banyan.api.audit import AuditAPI
from banyan.model.audit import AuditEvent, AuditAction, AuditEventType


class AuditController(Controller):
    class Meta:
        label = 'audit'
        stacked_type = 'nested'
        stacked_on = 'base'
        help = 'retrieve audit logs'

    @property
    def _client(self) -> AuditAPI:
        return self.app.client.audit

    @ex(help='list events',
        arguments=[
            (['--type'],
             {
                 'help': 'Filters by admin activity type.',
                 'choices': AuditEventType.choices(),
             }),
            (['--action'],
             {
                 'help': 'Filters events by one of the possible event action values.',
                 'choices': AuditAction.choices(),
             }),
            (['--admin-email'],
             {
                 'help': 'Filters by Admin email address.',
             }),
            (['--before'],
             {
                 'help': 'Filters events that occurred before a specific time.',
                 'type': datetime.fromisoformat,
             }),
            (['--after'],
             {
                 'help': 'Filters events that occurred after a specific time.',
                 'type': datetime.fromisoformat,
             }),
        ]
        )
    def list(self):
        events: List[AuditEvent] = self._client.list(before_dt=self.app.pargs.before, after_dt=self.app.pargs.after,
                                                     event_type=self.app.pargs.type, action=self.app.pargs.action,
                                                     admin_email=self.app.pargs.admin_email)
        event_json = AuditEvent.Schema().dump(events, many=True)
        self.app.render(event_json, handler='json', indent=2, sort_keys=True)



from datetime import datetime, timedelta
from typing import List, Dict, Any, Set
from uuid import UUID

from banyan.api.base import ServiceBase, Resource
from banyan.model import BanyanApiObject
from banyan.model.event_v2 import EventV2, EventOrID


class EventV2API(ServiceBase):
    class Meta:
        data_class = EventV2
        info_class = EventV2
        arg_type = EventOrID
        list_uri = '/events'
        uri_param = 'EventID'
        obj_name = 'event'
        supports_paging = True

    def create(self, obj: BanyanApiObject) -> str:
        raise NotImplementedError('The Banyan API does not support this operation')

    def update(self, obj: BanyanApiObject) -> str:
        raise NotImplementedError('The Banyan API does not support this operation')

    def delete(self, obj: BanyanApiObject) -> str:
        raise NotImplementedError('The Banyan API does not support this operation')

    def list2(self, before_dt: datetime = None, after_dt: datetime = None, order: str = None,
              event_type: str = None, subtype: str = None, action: str = None,
              email_address: str = None, device_id: str = None, device_serial: str = None,
              container_id: str = None, service_name: str = None, event_id: str = None) -> list:
        params = self._make_params(before_dt, after_dt, order, event_type, subtype, action, email_address,
                                   device_id, device_serial, container_id, service_name, event_id)
        return self._list_paged(params)

    def list(self, before_dt: datetime = None, after_dt: datetime = None, order: str = None,
             event_type: str = None, subtype: str = None, action: str = None,
             email_address: str = None, device_id: str = None, device_serial: str = None,
             container_id: str = None, service_name: str = None, event_id: str = None, limit: int = None) -> list:
        params = self._make_params(before_dt, after_dt, order, event_type, subtype, action, email_address,
                                   device_id, device_serial, container_id, service_name, event_id, limit)
        if before_dt and after_dt:
            return self._list_daterange(params)
        elif before_dt:
            params['after'] = int((datetime.now() - timedelta(days=7)).timestamp() * 1000)
            return self._list_daterange(params)
        elif after_dt:
            params['before'] = int(datetime.now().timestamp() * 1000)
            return self._list_daterange(params)
        else:
            return self._list_paged(params)

    def _list_daterange(self, params: Dict[str, Any]) -> List[Resource]:
        order = params.get('order', 'desc')
        params['order'] = 'asc'
        all_data: List[EventV2] = list()
        event_ids: Set[UUID] = set()
        while params['after'] < params['before']:
            response_json = self._client.api_request('GET', self.Meta.list_uri, params=params)
            data: List[EventV2] = self.Meta.info_class.Schema().load(response_json['data'], many=True)
            data = list(filter(lambda x: x.event_id not in event_ids, data))
            if len(data) == 0:
                break
            all_data.extend(data)
            event_ids.update([x.event_id for x in data])
            params['after'] = int(data[-1].created_at.timestamp() * 1000)
        self._build_cache(all_data)
        return all_data

    def _list_paged(self, params: Dict[str, Any]) -> List[Resource]:
        response_json = self._client.paged_request('GET', self.Meta.list_uri, params=params)
        data: List[EventV2] = self.Meta.info_class.Schema().load(response_json, many=True)
        self._build_cache(data)
        return data

    @staticmethod
    def _make_params(before_dt: datetime = None, after_dt: datetime = None, order: str = None,
                     event_type: str = None, subtype: str = None, action: str = None,
                     email_address: str = None, device_id: str = None, device_serial: str = None,
                     container_id: str = None, service_name: str = None, event_id: str = None,
                     limit: int = None) -> Dict[str, Any]:
        return ServiceBase.args_to_html_params([
            (before_dt, 'before', lambda: int(before_dt.timestamp() * 1000)),
            (after_dt, 'after', lambda: int(after_dt.timestamp() * 1000)),
            (order, 'order', order),
            (event_type, 'type', event_type),
            (subtype, 'sub_type', subtype),
            (action, 'action', action),
            (email_address, 'user_email', email_address),
            (device_id, 'device_id', device_id),
            (device_serial, 'serialnumber', device_serial),
            (container_id, 'workload_container_id', container_id),
            (service_name, 'service_name', service_name),
            (event_id, 'id', event_id),
            (limit, 'limit', limit),
        ])



import boto3
from ec2_metadata import ec2_metadata


class AwsCloud:
    def __init__(self):
        self._elb = boto3.client('elbv2')
        self._instance_id = ec2_metadata.instance_id

    def _has_banyanops_tag(self, elb_arn: str) -> bool:
        response = self._elb.describe_tags(ResourceArns=[elb_arn])
        assert ('TagDescriptions' in response)
        for tag_desc in response['TagDescriptions']:
            for tag in tag_desc['Tags']:
                if tag['Key'] == 'Provider' and tag['Value'] == 'BanyanOps':
                    return True
        return False

    def _get_target_group_443_arn(self, elb_arn: str) -> str:
        response = self._elb.describe_target_groups(LoadBalancerArn=elb_arn)
        assert ('TargetGroups' in response)
        for tg in response['TargetGroups']:
            if tg['Port'] == 443:
                return tg['TargetGroupArn']
        raise ValueError(f'ELB with ARN {elb_arn} does not have a target group on :443')

    # noinspection PyMethodMayBeStatic
    def _is_healthy_nlb(self, elb: dict) -> bool:
        return elb['Type'] == 'network' and elb['Scheme'] == 'internet-facing' and elb['State']['Code'] == 'active'

    def _am_i_a_target(self, tg_arn: str, instance_id: str) -> bool:
        response = self._elb.describe_target_health(TargetGroupArn=tg_arn)
        assert ('TargetHealthDescriptions' in response)
        for tg_health in response['TargetHealthDescriptions']:
            target_id = tg_health['Target']['Id']
            if target_id == instance_id:
                return True
        return False

    def get_public_dns_name(self) -> str:
        response = self._elb.describe_load_balancers()
        assert ('LoadBalancers' in response)
        for elb in response['LoadBalancers']:
            elb_arn = elb['LoadBalancerArn']
            elb_dns_name = elb['DNSName']
            if not self._is_healthy_nlb(elb):
                continue
            tg_arn = self._get_target_group_443_arn(elb_arn)
            if not self._am_i_a_target(tg_arn, self._instance_id):
                continue
            return elb_dns_name
        raise RuntimeError(f'could not find a public NLB for instance-id {self._instance_id}')



import json
from dataclasses import field
from datetime import datetime
from typing import ClassVar, Type, Optional
from uuid import UUID

from marshmallow import Schema, validate, pre_load, EXCLUDE
from marshmallow_dataclass import dataclass

from banyan.model import Resource, NanoTimestampField, BanyanEnum
from banyan.model.attachment import Attachment
from banyan.model.policy import PolicyInfo
from banyan.model.role import RoleInfo
from banyan.model.service import ServiceInfo


class AuditEventType(BanyanEnum):
    ADMIN_SIGN_ON = "admin_sign_on"
    IDP_SETTINGS = "idp_settings"
    MDM_SETTINGS = "mdm_settings"
    TRUSTSCORE_FACTORS = "trustscore_factors"
    UNKNOWN_DEVICE_SETTINGS = "unknown_device"
    SECURITY_ATTACH_POLICY = "security_attach_policy"
    POLICY = "policy"
    REGISTERED_SERVICE = "registered_service"
    ROLE = "role"
    DEVICE_REGISTRATION_IDP_SETTINGS = "device_registration_idp_settings"
    ADMIN_USER = "admin_user"
    MDM_DEPLOY_OTP_SKIP_ROLE = "mdm_deploy_otp_skip_role"
    MDM_DEPLOY_KEY = "mdm_deploy_key"
    INVITATION_CODE = "invitation_code"
    TRUSTSCORE_TTL = "trustscore_ttl"
    ENDUSER_DEVICE = "enduser_device"
    PREFERRED_APPLICATIONS = "preferred_applications"
    LATEST_OS_CONFIG = "latest_os_config"
    TRUST_CONFIG = "trust_config"
    ROOT_CERTS = "root_certs"
    SAAS_APPLICATIONS = "saas_applications"
    IDP_ROUTED_APPLICATIONS = "idp_routed_applications"


class AuditAction(BanyanEnum):
    UPDATE = "update"
    CREATE = "create"
    DELETE = "delete"
    ENABLE = "enable"
    DISABLE = "disable"


@dataclass
class AuditEvent(Resource):
    class Meta:
        unknown = EXCLUDE

    event_id: str = field(metadata={"data_key": "id"})
    event_type: str = field(metadata={"data_key": "type", "validate": validate.OneOf(AuditEventType.choices())})
    org_id: UUID
    created_at: datetime = field(metadata={"marshmallow_field": NanoTimestampField()})
    message: str
    action: str = field(metadata={"validate": validate.OneOf(AuditAction.choices())})
    admin_email: str
    changes_new: Optional[str]
    changes_old: Optional[str]
    Schema: ClassVar[Type[Schema]] = Schema

    @property
    def name(self) -> str:
        return self.id

    @property
    def id(self) -> str:
        return str(self.event_id)

    def _extract_obj(self, obj_str: str) -> Resource:
        if self.event_type == AuditEventType.REGISTERED_SERVICE:
            return ServiceInfo.Schema().loads(obj_str)
        elif self.event_type == AuditEventType.POLICY:
            return PolicyInfo.Schema().loads(obj_str)
        elif self.event_type == AuditEventType.ROLE:
            return RoleInfo.Schema().loads(obj_str)
        elif self.event_type == AuditEventType.SECURITY_ATTACH_POLICY:
            return Attachment.Schema().loads(obj_str)
        else:
            raise ValueError(f'unknown object type for deserialization: {self.event_type}')

    @property
    def object_old(self) -> Resource:
        return self._extract_obj(self.changes_old) if self.changes_old else None

    @property
    def object_new(self) -> Resource:
        return self._extract_obj(self.changes_new) if self.changes_new else None

    # noinspection PyUnusedLocal
    # pylint: disable=W0613,R0201
    @pre_load
    def _remove_empty_dates(self, data, many, **kwargs):
        if "changes_new" in data and data["changes_new"] is not None:
            data["changes_new"] = json.dumps(data["changes_new"])
        if "changes_old" in data and data["changes_old"] is not None:
            data["changes_old"] = json.dumps(data["changes_old"])
        return data



from dataclasses import field
from datetime import datetime
from ipaddress import IPv4Address
from typing import List, Dict, ClassVar, Type
from uuid import UUID

import OpenSSL.crypto
from marshmallow import fields, Schema, EXCLUDE
from marshmallow_dataclass import dataclass
from semver import VersionInfo

from banyan.model import Resource, VersionField
from banyan.model.netagent import Netagent


@dataclass
class Shield(Resource):
    class Meta:
        unknown = EXCLUDE

    org_id: UUID = field(metadata={"data_key": "OrgID"})
    shield_id: UUID = field(metadata={"data_key": "UUID"})
    shield_name: str = field(metadata={"data_key": "ShieldName"})
    group_type: str = field(metadata={"data_key": "GroupType"})
    cluster_mgr_type: str = field(metadata={"data_key": "ClusterMgrType"})
    cluster_mgr_ip: IPv4Address = field(metadata={"marshmallow_field": fields.String(data_key="ClusterMgrIP")})
    version: VersionInfo = field(metadata={"marshmallow_field": VersionField(data_key="ShieldVersion")})
    auto_upgrade: bool = field(metadata={"data_key": "AutoUpgrade"})
    one_time_key: UUID = field(metadata={"data_key": "OneTimeKey"})
    otk_expire_dt: datetime = field(metadata={"data_key": "OTKExpiryTime"})
    scep_enabled: bool = field(metadata={"data_key": "SCEPEnabled"})
    public_addr: str = field(metadata={"data_key": "PublicAddr"})
    ca_cert_str: str = field(metadata={"data_key": "CACert"})
    otk_enabled: bool = field(metadata={"data_key": "OTKEnabled"})
    ssh_ca_public_key: str = field(metadata={"data_key": "SSHCAPublicKey"})
    Schema: ClassVar[Type[Schema]] = Schema

    def load_cacert(self) -> OpenSSL.crypto.X509:
        cert = OpenSSL.crypto.load_certificate(OpenSSL.crypto.FILETYPE_PEM, self.ca_cert_str.encode('utf-8'))
        return cert

    @property
    def name(self) -> str:
        return self.shield_name

    @property
    def id(self) -> str:
        return str(self.shield_id)


@dataclass
class ShieldLastActivity:
    class Meta:
        unknown = EXCLUDE

    last_activity_time: datetime = field(metadata={"data_key": "InsertTime"})
    status: str = field(metadata={"data_key": "Status"})
    uuid: UUID = field(metadata={"data_key": "UUID"})
    Schema: ClassVar[Type[Schema]] = Schema


@dataclass
class ShieldConfig:
    class Meta:
        unknown = EXCLUDE

    shields: List[Shield] = field(default_factory=list, metadata={"data_key": "Configs"})
    last_activity_map: Dict[str, ShieldLastActivity] = field(default_factory=dict,
                                                             metadata={"data_key": "ShieldLastActivitiesMap"})
    netagent_map: Dict[str, List[Netagent]] = field(default_factory=dict,
                                                    metadata={"data_key": "NetagentHostInfoMap"})
    Schema: ClassVar[Type[Schema]] = Schema



from dataclasses import field
from datetime import datetime
from typing import List, ClassVar, Type, Optional, Union
from uuid import UUID

from marshmallow import Schema, validate, pre_load, EXCLUDE
from marshmallow.fields import String
from marshmallow_dataclass import dataclass

from banyan.model import Resource, NanoTimestampField, MilliTimestampField, BanyanEnum, TimestampField


@dataclass
class EventUser:
    class Meta:
        unknown = EXCLUDE

    email: str
    groups: List[str] = field(default_factory=list)
    roles: List[str] = field(default_factory=list)

    # noinspection PyUnusedLocal
    @pre_load
    def _remove_nulls(self, data, many, **kwargs):
        if "roles" in data and data["roles"] is None:
            del data["roles"]
        if "groups" in data and data["groups"] is None:
            del data["groups"]
        return data


@dataclass
class EventDevice:
    class Meta:
        unknown = EXCLUDE

    device_id: str = field(metadata={"data_key": "id"})
    friendly_name: str
    mac_address: str
    serial_number: str
    unregistered: bool
    compromised_status: str
    compliance_status: str
    oem_info: str
    model: str
    platform: str
    ownership: str
    architecture: str
    udid: str
    source: str
    last_mdm_data_synced_at: datetime = field(metadata={'marshmallow_field': TimestampField()})
    last_user_agent: Optional[str]


@dataclass
class EventClient:
    class Meta:
        unknown = EXCLUDE

    user_agent: str
    ip_address: str


@dataclass
class EventUserPrincipal:
    class Meta:
        unknown = EXCLUDE

    user: Optional[EventUser]
    device: Optional[EventDevice]
    client: Optional[EventClient]


@dataclass
class EventWorkloadSet:
    class Meta:
        unknown = EXCLUDE

    container_name: str
    container_id: str
    image: str
    repo: str
    tag: str
    labels: str
    container_ips: str
    app_name: str


@dataclass
class EventWorkloadHost:
    class Meta:
        unknown = EXCLUDE

    host_ips: str
    host_name: str
    cluster_id: str
    port_map: str


@dataclass
class EventWorkloadPrincipal:
    class Meta:
        unknown = EXCLUDE

    workload_set: EventWorkloadSet
    host: EventWorkloadHost


@dataclass
class EventRoleInfo:
    class Meta:
        unknown = EXCLUDE

    role_id: UUID = field(metadata={"data_key": "id"})
    role_name: str = field(metadata={"data_key": "name"})
    version: int
    bound_by: str
    bound_at: datetime = field(metadata={'marshmallow_field': NanoTimestampField()})


@dataclass
class EventPolicyInfo:
    class Meta:
        unknown = EXCLUDE

    policy_id: str = field(metadata={"data_key": "id"})
    policy_name: str = field(metadata={"data_key": "name"})
    version: int
    attached_by: str
    attached_at: datetime = field(metadata={'marshmallow_field': MilliTimestampField()})
    enabled: bool


@dataclass
class EventServiceInfo:
    class Meta:
        unknown = EXCLUDE

    service_id: str = field(metadata={"data_key": "id"})
    service_name: str = field(metadata={"data_key": "name"})
    service_type: str = field(metadata={"data_key": "type"})
    version: int = field(metadata={'marshmallow_field': String()})


@dataclass
class EventAccessLevel:
    class Meta:
        unknown = EXCLUDE

    resource: str


@dataclass
class EventSniData:
    class Meta:
        unknown = EXCLUDE

    name_requested: str
    name_matched: str


@dataclass
class EventRequestData:
    class Meta:
        unknown = EXCLUDE

    protocol: str
    request_type: str = field(metadata={"data_key": "type"})
    query_crud_types: str
    query_resources: str


@dataclass
class EventChannelInfo:
    class Meta:
        unknown = EXCLUDE

    access_level: EventAccessLevel
    sni_data: EventSniData
    request_data: EventRequestData


@dataclass
class EventLinkSource:
    class Meta:
        unknown = EXCLUDE

    container_id: str
    container_name: str
    service_id: str
    service_name: str
    service_version: Optional[int] = field(metadata={"marshmallow_field": String()})
    host_name: str
    ip: str


@dataclass
class EventLinkDestination:
    class Meta:
        unknown = EXCLUDE

    container_id: str
    container_name: str
    service_id: str
    service_name: str
    service_version: Optional[int] = field(metadata={"marshmallow_field": String()})
    host_name: str
    ip: str
    port: Optional[str]


@dataclass
class EventLinkInfo:
    class Meta:
        unknown = EXCLUDE

    source: EventLinkSource
    destination: EventLinkDestination


@dataclass
class EventTrustScore:
    class Meta:
        unknown = EXCLUDE

    trustscore_id: str = field(metadata={"data_key": "id"})
    trustscore_type: str = field(metadata={"data_key": "type"})
    timestamp: datetime = field(metadata={'marshmallow_field': NanoTimestampField()})
    score: int


class EventV2Severity(BanyanEnum):
    DEBUG = "DEBUG"
    INFO = "INFO"
    WARN = "WARN"
    ERROR = "ERROR"


class EventV2Type(BanyanEnum):
    ACCESS = "Access"
    IDENTITY = "Identity"
    REGISTRATION = "Registration"
    TRUST_SCORING = "TrustScoring"


class EventV2Subtype(BanyanEnum):
    DEVICE = "Device"
    USER_PRINCIPAL = "UserPrincipal"
    CONNECTION = "Connection"
    RESOURCE = "Resource"


@dataclass
class EventV2(Resource):
    class Meta:
        unknown = EXCLUDE

    event_id: UUID = field(metadata={"data_key": "id"})
    external_id: Optional[str]
    org_id: UUID
    org_name: str
    event_type: str = field(metadata={"data_key": "type", "validate": validate.OneOf(EventV2Type.choices())})
    sub_type: str = field(metadata={"validate": validate.OneOf(EventV2Subtype.choices())})
    severity: str = field(metadata={"validate": validate.OneOf(EventV2Severity.choices())})
    action: str
    message: str
    result: str
    created_at: datetime = field(metadata={'marshmallow_field': MilliTimestampField()})
    user_principal: EventUserPrincipal
    workload_principal: Optional[EventWorkloadPrincipal]
    role: Optional[List[EventRoleInfo]]
    trustscore: Optional[EventTrustScore]
    service: Optional[EventServiceInfo]
    policy: Optional[EventPolicyInfo]
    channel: Optional[EventChannelInfo]
    link: Optional[EventLinkInfo]
    Schema: ClassVar[Type[Schema]] = Schema

    @property
    def name(self) -> str:
        return self.id

    @property
    def id(self) -> str:
        return str(self.event_id)


@dataclass
class EventV2TypeCount(Resource):
    class Meta:
        unknown = EXCLUDE

        end_date: datetime = field(metadata={"data_key": "StatsEndTime"})
        delta_time: int = field(metadata={"data_key": "DeltaTime"})
        access_auth: int = field(metadata={"data_key": "NumEventTypeAccessAuth"})
        access_unauth: int = field(metadata={"data_key": "NumEventTypeAccessUnauth"})
        identity_grant: int = field(metadata={"data_key": "NumEventTypeIdentityGrant"})
        identity_deny: int = field(metadata={"data_key": "NumEventTypeIdentityDeny"})
        service: int = field(metadata={"data_key": "NumEventTypeService"})
        link: int = field(metadata={"data_key": "NumEventTypeLink"})
        policy: int = field(metadata={"data_key": "NumEventTypePolicy"})
        role: int = field(metadata={"data_key": "NumEventTypeRole"})


EventOrID = Union[EventV2, str]



from datetime import datetime
from typing import List
import os

from cement import Controller, ex

from banyan.api import BanyanApiClient
from banyan.model.service import ServiceInfo


class ExportController(Controller):
    class Meta:
        label = 'export'
        stacked_type = 'nested'
        stacked_on = 'base'
        help = 'export all objects from an organization'

    @property
    def _client(self) -> BanyanApiClient:
        return self.app.client

    @ex(help='export all objects from an organization',
        arguments=[
            (['--path'],
             {
                 'help': 'Directory to place the object JSON files.',
                 'default': os.getcwd()
             }),
        ]
        )
    def all(self):
        self.roles()
        self.policies()
        self.services()

    @ex(help='export roles from an organization',
        arguments=[
            (['--path'],
             {
                 'help': 'Directory to place the object JSON files.',
                 'default': os.getcwd()
             }),
        ]
        )
    def roles(self):
        role_path = self._mkdir('roles')
        self.app.print(f' * Exporting roles to {role_path}')
        all_roles = self._client.roles.list()
        for role in all_roles:
            with open(self._mkfile(role_path, f'{role.name}.role.json'), 'w') as f:
                f.write(role.role.Schema().dumps(role.role))

    @ex(help='export policies from an organization',
        arguments=[
            (['--path'],
             {
                 'help': 'Directory to place the object JSON files.',
                 'default': os.getcwd()
             }),
        ]
        )
    def policies(self):
        policy_path = self._mkdir('policies')
        self.app.print(f' * Exporting policies to {policy_path}')
        all_policies = self._client.policies.list()
        for policy in all_policies:
            with open(self._mkfile(policy_path, f'{policy.name}.policy.json'), 'w') as f:
                f.write(policy.policy.Schema().dumps(policy.policy))

    @ex(help='export services from an organization',
        arguments=[
            (['--path'],
             {
                 'help': 'Directory to place the object JSON files.',
                 'default': os.getcwd()
             }),
        ]
        )
    def services(self):
        service_path = self._mkdir('services')
        self.app.print(f' * Exporting services to {service_path}')
        all_services = self._client.services.list()
        for service in all_services:
            with open(self._mkfile(service_path, f'{service.name}.service.json'), 'w') as f:
                f.write(service.service.Schema().dumps(service.service))

    def _mkdir(self, subdir: str) -> str:
        path = os.path.join(self.app.pargs.path, subdir)
        os.makedirs(path, exist_ok=True)
        return os.path.abspath(path)

    def _mkfile(self, path: str, fname: str) -> str:
        fname = fname.replace(' ', '_')
        return os.path.join(path, fname)



from datetime import datetime
from typing import List

from banyan.api.base import ServiceBase, Resource
from banyan.model import BanyanApiObject
from banyan.model.audit import AuditEvent


class AuditAPI(ServiceBase):
    class Meta:
        data_class = AuditEvent
        info_class = AuditEvent
        arg_type = str
        list_uri = '/audit_logs'
        uri_param = ''
        obj_name = 'audit log'
        supports_paging = True

    def create(self, obj: BanyanApiObject) -> str:
        raise NotImplementedError('The Banyan API does not support this operation')

    def update(self, obj: BanyanApiObject) -> str:
        raise NotImplementedError('The Banyan API does not support this operation')

    def delete(self, obj: BanyanApiObject) -> str:
        raise NotImplementedError('The Banyan API does not support this operation')

    def list(self, before_dt: datetime = None, after_dt: datetime = None,
             event_type: str = None, action: str = None, admin_email: str = None) -> list:
        params = ServiceBase.args_to_html_params([
            (before_dt, 'end_time', int(before_dt.timestamp() * 1000000000) if before_dt else None),
            (after_dt, 'start_time', int(after_dt.timestamp() * 1000000000) if after_dt else None),
            (event_type, 'type', event_type),
            (action, 'action', action),
            (admin_email, 'admin_email', admin_email),
        ])
        response_json = self._client.api_request('GET', self.Meta.list_uri, params=params)
        data: List[Resource] = self.Meta.info_class.Schema().load(response_json, many=True)
        self._build_cache(data)
        return data



import os
import re
import socket
import textwrap
import urllib.parse
from collections import namedtuple
from dataclasses import dataclass, field
from typing import Optional, List

import dns.resolver
import requests
import requests.exceptions
from colorama import Style, Fore

from banyan.api import BanyanApiClient
from banyan.model.service import Service, BackendTarget
from banyan.model.netagent import Netagent
from banyan.lib.cloud import get_cloud_vendor
from banyan.lib.cloud.aws import AwsCloud
from banyan.lib.certs import CertPair

DEFAULT_TIMEOUT = 0.5
DEFAULT_DNS_SERVER = '8.8.8.8'


@dataclass
class TestResult:
    passed: bool
    details: str = ''
    info: List[str] = field(default_factory=list)


class ServiceTest:
    def __init__(self, service: Service, client: BanyanApiClient, timeout: float = DEFAULT_TIMEOUT,
                 external_dns_server: str = DEFAULT_DNS_SERVER,
                 wildcard_hostname: Optional[str] = None):
        self._service = service
        self._client = client
        self._timeout = timeout
        self._ext_dns = external_dns_server
        self._wildcard = wildcard_hostname
        self._my_hostname = socket.gethostname()
        self._agent = self._find_access_tier()
        self._vendor = get_cloud_vendor()
        self._cloud = self._get_cloud()
        self._dns_name = self._cloud.get_public_dns_name()

        self._resolver = dns.resolver.Resolver(configure=False)
        self._resolver.nameservers = [self._ext_dns]
        answers = self._resolver.resolve(self._dns_name, 'A', search=False)
        self._external_ips = [x.address for x in answers]

    def _find_access_tier(self) -> Netagent:
        for agent in self._client.netagents.list():  # type: Netagent
            if agent.hostname == self._my_hostname:
                return agent
        raise RuntimeError(f'Unable to find a netagent with hostname {self._my_hostname}')

    def _get_cloud(self) -> AwsCloud:
        if self._vendor == 'aws':
            return AwsCloud()
        else:
            raise RuntimeError(f'Cloud vendor "{self._vendor} is not supported')

    @property
    def _cert_dns_names(self) -> List[str]:
        return self._service.spec.cert_settings.dns_names

    @property
    def _dns_names(self) -> List[str]:
        return self._cert_dns_names or [self._service.metadata.tags.domain]

    @property
    def _target(self) -> BackendTarget:
        return self._service.spec.backend.target

    @staticmethod
    def _multimatch(value: str, multi_string: str) -> bool:
        return any([x.strip() == value for x in multi_string.split('|')])

    def _backend_hostname(self) -> str:
        host = self._target.name
        template_pattern = re.compile('{{([^}]*)}}')
        while '{{' in host:
            if not self._wildcard:
                raise RuntimeError(
                    'Service target contains a template. You must specify a value for the --wildcard option.')
            match = template_pattern.search(host)
            template_tokens = match.group(1).strip().split()
            if template_tokens[0] == '.Name':
                host = host.replace(match.group(0), self._wildcard)
            elif template_tokens[0] == 'index' and template_tokens[1] == '.Parts':
                index = int(template_tokens[2])
                host_tokens = self._wildcard.split(self._target.name_delimiter)
                host = host.replace(match.group(0), host_tokens[index])
        if host in self._service.spec.backend.dns_overrides.keys():
            host = self._service.spec.backend.dns_overrides[host]
        return host

    def run(self):
        print(f'Testing service {self._service.name} on host {self._my_hostname}:\n')
        print(f'Cloud vendor:        {self._vendor}')
        print(f'External hostname:   {self._dns_name}')
        print(f'External IPs:        {self._external_ips}')
        print()
        test_methods = [x for x in dir(self) if x.startswith('_test_')]
        for method_name in sorted(test_methods):
            method = getattr(self, method_name)
            docstring = method.__doc__
            print(docstring, end='... ')
            result: TestResult = method()
            print(f'{Fore.GREEN}PASSED' if result.passed else f'{Fore.RED}FAILED', end='')
            print(Style.RESET_ALL)
            if not result.passed:
                print(Fore.RED, end='')
                if result.details:
                    print(result.details)
                for info in result.info:
                    print(f'  * {info}')
                print(Style.RESET_ALL, end='')

    def _test_10_service_applies_here(self) -> TestResult:
        """Selected service applies to this netagent"""
        result = TestResult(True)
        matches = list()
        for i, selector in enumerate(self._service.spec.attributes.host_tag_selector):
            match = True
            for key, value in selector.items():
                if key in self._agent.host_tags.keys() and not self._multimatch(self._agent.host_tags[key], value):
                    match = False
                    result.info.append(
                        f'Selector {i}: {key}: {value} (expected) != {self._agent.host_tags[key]} (actual)')
            matches.append(match)
        if not any(matches):
            result.passed = False
            result.details = 'This netagent does not match any of the host tag selectors defined in the service ' \
                             'spec. Further results from this test may not be meaningful.'
        return result

    def _test_20_dns_names_match_tls_sni(self) -> TestResult:
        """DNS names in certificate match TLS SNI settings"""
        cert_names = set(self._service.spec.cert_settings.dns_names)
        sni_names = set(self._service.spec.attributes.tls_sni)
        if cert_names == sni_names:
            return TestResult(True)
        else:
            return TestResult(False, f'The DNS names in the service spec {list(cert_names)} do not match '
                                     f'the values specified for the TLS SNI header {list(sni_names)}. '
                                     f'Normally these should be the same.')

    def _test_30_meta_domain_matches(self) -> TestResult:
        """Meta domain matches DNS and TLS SNI settings"""
        meta_name = self._service.metadata.tags.domain
        cert_names = self._service.spec.cert_settings.dns_names
        sni_names = self._service.spec.attributes.tls_sni
        if meta_name in cert_names and meta_name in sni_names:
            return TestResult(True)
        else:
            return TestResult(False, f'The domain name "{meta_name}" in the service metadata was missing from '
                                     f'either the DNS names of the service spec {cert_names} or the TLS SNI '
                                     f'header values {sni_names}.')

    def _test_40_meta_port_in_ports(self) -> TestResult:
        """Meta port matches frontend listener port configuration"""
        meta_port = int(self._service.metadata.tags.port)
        frontend_ports = [int(x.port) for x in self._service.spec.attributes.frontend_addresses]
        if meta_port in frontend_ports:
            return TestResult(True)
        else:
            return TestResult(False, f'The port number in the service metadata ({meta_port}) does not match '
                                     f'any of the frontend ports for the service ({frontend_ports}).')

    def _test_50_public_dns(self) -> TestResult:
        """Public DNS records for service name resolve to this Access Tier"""
        result = TestResult(True)
        for name in self._dns_names:
            try:
                answers = self._resolver.resolve(name, 'A', search=False)
                dns_results = [x.address for x in answers]
                if set(dns_results) != set(self._external_ips):
                    result.passed = False
                    result.info.append(f'Public DNS for {name} resolved to {dns_results} '
                                       f'instead of {self._external_ips}')
            except dns.resolver.NXDOMAIN:
                result.passed = False
                result.info.append(f'Public DNS query against {self._resolver.nameservers} for {name} '
                                   f'did not return any records')
        if not result.passed:
            result.details = f'One or more of the DNS names specified for this service ({self._dns_names}) ' \
                             f'does not have a public DNS record pointing to the IP addresses of this ' \
                             f'Access Tier ({self._external_ips})'
        return result

    def _test_60_custom_tls_cert(self) -> TestResult:
        """Custom TLS cert exists (if specified) and is correct"""
        result = TestResult(True)
        custom_tls = self._service.spec.cert_settings.custom_tls_cert
        if custom_tls.enabled:
            try:
                cert = CertPair(custom_tls.cert_file, custom_tls.key_file)
                for name in self._dns_names:
                    if not cert.is_name_match(name):
                        result.passed = False
                        result.details = 'The path points to a valid certificate, but the DNS names of the ' \
                                         'certificate do not match the DNS names configured for the service.'
                        result.info.append(f'Service DNS name {name} does not match names configured in '
                                           f'the certificate ({cert.get_names()})')
            except FileNotFoundError:
                result.passed = False
                result.details = f'One of the filenames configured for the custom certificate could not be found.'
                if not os.path.exists(custom_tls.cert_file):
                    result.info.append(f'Certificate file {custom_tls.cert_file} is missing')
                if not os.path.exists(custom_tls.key_file):
                    result.info.append(f'Private key file {custom_tls.key_file} is missing')
            except PermissionError:
                result.passed = False
                result.details = 'The path points to a valid certificate, but the certificate could not be read ' \
                                 'due to the file permissions.'
        return result

    def _tcp_connect(self) -> TestResult:
        host = self._backend_hostname()
        result = TestResult(True)
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.settimeout(self._timeout)
        try:
            s.connect((host, int(self._target.port)))
        except socket.timeout:
            result.passed = False
            result.details = f'Connection to backend host {host} on port {self._target.port} timed out ' \
                             f'after {self._timeout} seconds.'
        finally:
            s.close()
        return result

    def _http_connect(self) -> TestResult:
        host = self._backend_hostname()
        result = TestResult(True)
        scheme = 'https' if self._target.tls else 'http'
        url = urllib.parse.urlunparse((scheme, f'{host}:{self._target.port}', '', '', '', ''))
        params = dict()
        if self._target.client_certificate:
            params['cert'] = ('netagent.crt', 'netagent.key')
        if self._target.tls_insecure:
            params['verify'] = False
        try:
            response = requests.get(url, timeout=self._timeout, **params)
            if response.status_code != 200:
                result.passed = False
                result.details = f'{scheme.upper()} connection to {host}:{self._target.port} succeeded, but the ' \
                                 f'request resulted in a status code {response.status_code}: {response.reason}.'
        except requests.exceptions.ConnectTimeout:
            result.passed = False
            result.details = f'{scheme.upper()} connection to {host}:{self._target.port} timed out ' \
                             f'after {self._timeout} seconds.'
        except requests.exceptions.ConnectionError as ex:
            result.passed = False
            result.details = f'{scheme.upper()} connection to {host}:{self._target.port} failed: {ex}.'
        return result

    def _test_70_backend_dns(self) -> TestResult:
        """Internal DNS records exist for backend host"""
        result = TestResult(True)
        name = self._backend_hostname()
        try:
            dns.resolver.resolve(name, 'A', search=True)
        except dns.resolver.NXDOMAIN:
            result.passed = False
            result.info.append(f'DNS for {name} did not return any records')
            result.details = f'Internal DNS for {name} did not resolve to any address records.'
        return result

    def _test_80_backend_connect(self) -> TestResult:
        """Connection to backend service"""
        if self._service.spec.http_settings.enabled:
            return self._http_connect()
        else:
            return self._tcp_connect()
